--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

----------------------------------------------------------------------------------
-- 			Includes code by Semlar 			--
----------------------------------------------------------------------------------

-- Some math stuff
local cos, sin, pi2, halfpi = math.cos, math.sin, math.rad(360), math.rad(90)
local function Transform(tx, x, y, angle, aspect) -- Translates texture to x, y and rotates about its center
	local c, s = cos(angle), sin(angle)
	local y, oy = y / aspect, 0.5 / aspect
	local ULx, ULy = 0.5 + (x - 0.5) * c - (y - oy) * s, (oy + (y - oy) * c + (x - 0.5) * s) * aspect
	local LLx, LLy = 0.5 + (x - 0.5) * c - (y + oy) * s, (oy + (y + oy) * c + (x - 0.5) * s) * aspect
	local URx, URy = 0.5 + (x + 0.5) * c - (y - oy) * s, (oy + (y - oy) * c + (x + 0.5) * s) * aspect
	local LRx, LRy = 0.5 + (x + 0.5) * c - (y + oy) * s, (oy + (y + oy) * c + (x + 0.5) * s) * aspect
	tx:SetTexCoord(ULx, ULy, LLx, LLy, URx, URy, LRx, LRy)
end

-- Frame functions 
local function OnSizeChanged(self, width, height)
	self._wedge:SetSize(width, height) -- it's important to keep this texture sized correctly
	self._aspect = width / height -- required to calculate the texture coordinates
end

local function SetClockwise(self, clockwise)
	self._clockwise = clockwise
end

local function SetFont(self, font, fontsize, fontflags)
	self.timerFrame.timer:SetFont(font, fontsize, fontflags)
end

local function SetReverse(self, reverse)
	self._reverse = reverse
end

local function SetTextShown(self, show)
	self.timerFrame.timer:SetShown(show)
end

local function SetValue(self, value)
	-- Correct invalid ranges, preferably just don't feed it invalid numbers
	if value > 1 then 
		value = 1
	elseif value < 0 then 
		value = 0 
	end
	-- Reverse our normal behavior
	if self._reverse then
		value = 1 - value
	end
	-- Determine which quadrant we're in
	local q, quadrant = self._clockwise and (1 - value) or value -- 4 - floor(value / 0.25)
	if q >= 0.75 then
		quadrant = 1
	elseif q >= 0.5 then
		quadrant = 2
	elseif q >= 0.25 then
		quadrant = 3
	else
		quadrant = 4
	end
	if self._quadrant ~= quadrant then
		self._quadrant = quadrant
		-- Show/hide necessary textures if we need to
		if self._clockwise then
			for i = 1, 4 do
				self._textures[i]:SetShown(i < quadrant)
			end
		else
			for i = 1, 4 do
				self._textures[i]:SetShown(i > quadrant)
			end
		end
		-- Move scrollframe/wedge to the proper quadrant
		self._scrollframe:Hide()
		self._scrollframe:SetAllPoints(self._textures[quadrant])
		self._scrollframe:Show()
	end
	-- Rotate the things
	local rads = value * pi2
	if not self._clockwise then 
		rads = -rads + halfpi 
	end
	Transform(self._wedge, -0.5, -0.5, rads, self._aspect)
	self._rtx:SetDuration(0)
	self._rtx:SetEndDelay(2147483647)
	self._rtx:SetOrigin('BOTTOMRIGHT', 0, 0)
	self._rtx:SetRadians(-rads)
	self._group:Play()
end

local function StartPlay(self, start, duration, enable)
	if self.NoPlay or (self.start and self.start == start) or not enable or enable == 0 or start <= 0 or duration <= 0 then 
		return 
	end
	self.duration = duration
	self.totalElapsed = GetTime() - start
	self.start = start
	self.timerFrame.timer:SetTextColor(1, 1, 0)
	self.timerFrame.timer:SetScale(1)
	self:Show()
end


-- Teture functions
-- Creates a function that calls a method on all textures at once
local function CreateTextureFunction(func, self, ...)
	return function(self, ...)
		for i = 1, 4 do
			local tx = self._textures[i]
			tx[func](tx, ...)
		end
		self._wedge[func](self._wedge, ...)
	end
end

-- Pass calls to these functions on our frame to its textures
local TextureFunctions = {
	SetTexture = CreateTextureFunction('SetTexture'),
	SetBlendMode = CreateTextureFunction('SetBlendMode'),
	SetVertexColor = CreateTextureFunction('SetVertexColor'),
	SetAlpha = CreateTextureFunction('SetAlpha'),
}

-- Get Started
function NS:CreatePrecious(parent, name, hidetext)
	local f = CreateFrame("Frame", name, parent)
	f:Hide()
--	f:SetAllPoints(parent)
	f:SetPoint("TOPLEFT", -3, 3)
	f:SetPoint("BOTTOMRIGHT", 3, -3)
	f.timerFrame = CreateFrame("Frame", nil, f)
	f.timerFrame:SetFrameLevel(f:GetFrameLevel()+2) 
	f.timerFrame:SetAllPoints()
	f.timerFrame.timer = f.timerFrame:CreateFontString()
	f.timerFrame.timer:SetFontObject(GameFontHighlight)
	f.timerFrame.timer:SetPoint("CENTER")
	f.timerFrame.timer:SetJustifyH("CENTER")
--	f.timerFrame.timer:SetJustifyV("CENTER") 
	f.timerFrame.timer:SetJustifyV("MIDDLE")
	f.timerFrame.timer:SetDrawLayer("BORDER")
	-- ScrollFrame clips the actively animating portion of the cooldown
	local scrollframe = CreateFrame("ScrollFrame", nil, f)
	scrollframe:SetPoint('BOTTOMLEFT', f, 'CENTER')
	scrollframe:SetPoint('TOPRIGHT')
	f._scrollframe = scrollframe

	local scrollchild = CreateFrame("Frame", nil, scrollframe)
	scrollframe:SetScrollChild(scrollchild)
	scrollchild:SetAllPoints(scrollframe)
	
	-- Wedge thing
	local wedge = scrollchild:CreateTexture()
	wedge:SetPoint('BOTTOMRIGHT', f, 'CENTER')
	f._wedge = wedge

	-- Top Right
	local trTexture = f:CreateTexture()
	trTexture:SetPoint('BOTTOMLEFT', f, 'CENTER')
	trTexture:SetPoint('TOPRIGHT')
	trTexture:SetTexCoord(0.5, 1, 0, 0.5)

	-- Bottom Right
	local brTexture = f:CreateTexture()
	brTexture:SetPoint('TOPLEFT', f, 'CENTER')
	brTexture:SetPoint('BOTTOMRIGHT')
	brTexture:SetTexCoord(0.5, 1, 0.5, 1)

	-- Bottom Left
	local blTexture = f:CreateTexture()
	blTexture:SetPoint('TOPRIGHT', f, 'CENTER')
	blTexture:SetPoint('BOTTOMLEFT')
	blTexture:SetTexCoord(0, 0.5, 0.5, 1)

	-- Top Left
	local tlTexture = f:CreateTexture()
	tlTexture:SetPoint('BOTTOMRIGHT', f, 'CENTER')
	tlTexture:SetPoint('TOPLEFT')
	tlTexture:SetTexCoord(0, 0.5, 0, 0.5)

	-- /4|1\ -- Clockwise texture arrangement
	-- \3|2/ --

	f._textures = {trTexture, brTexture, blTexture, tlTexture}
	f._quadrant = nil -- Current active quadrant
	f._clockwise = true -- fill clockwise
	f._reverse = false -- Treat the provided value as its inverse, eg. 75% will display as 25%
	f._aspect = 1 -- aspect ratio, width / height of cooldown frame
	f:HookScript('OnSizeChanged', OnSizeChanged)

	for method, func in pairs(TextureFunctions) do
		f[method] = func
	end

	f.SetClockwise = SetClockwise
	f.SetReverse = SetReverse

	local group = wedge:CreateAnimationGroup()
	group:SetScript('OnFinished', function() group:Play() end)
	local rotation = group:CreateAnimation('Rotation')
	f._rtx = rotation
	f._group = group

	f:SetClockwise(false)
	f:SetReverse(true)

	f.totalElapsed = 0
	f.duration = 0
	f:SetScript('OnUpdate', function(self, elapsed)
			if self.start > GetTime() then return end
			self.totalElapsed = self.totalElapsed + elapsed
			if self.totalElapsed >= self.duration then
			    self.totalElapsed = 0
			    self.duration = 0
			    self:Hide()
			end
			local value
			if self.duration == 0 then
				value = 0
			else 
				value = self.totalElapsed / self.duration
			end
			local cdTime, LastLegs = NS:TimeLeft(self.totalElapsed, self.duration, true, 5)
			self.timerFrame.timer:SetText(cdTime)
			if LastLegs then
				self.timerFrame.timer:SetTextColor(1, 0, 0)
				self.timerFrame.timer:SetScale(1.7)
			end
			SetValue(self, value)
		end)
	f.StartPlay = StartPlay
	f.OnSizeChanged = OnSizeChanged
	f.SetClockwise = SetClockwise
	f.SetFont = SetFont
	f.SetReverse = SetReverse
	f.SetTextShown = SetTextShown
	SetTextShown(f, not hidetext)
	return f
end
